# -*- coding: utf-8 -*-
# LysKOM Protocol A version 10/11 client interface for Python
# (C) 1999-2002 Kent Engström. Released under GPL.
# (C) 2008 Henrik Rindlöw. Released under GPL.
# (C) 2012-2014 Oskar Skoog. Released under GPL.

from .datatypes import (
    ArrayAuxItem,
    String,
    ConfNo,
    TextNo,
    PersNo,
    Int32,
    TextStat,
    SessionNo,
    WhoInfo)


class AsyncMessages(object):
    """Enum like object which give names to the async message numbers
    in the protocol.
    """
    NEW_TEXT_OLD = 0
    NEW_NAME = 5
    I_AM_ON = 6
    SYNC_DB = 7
    LEAVE_CONF = 8
    LOGIN = 9
    REJECTED_CONNECTION = 11
    SEND_MESSAGE = 12
    LOGOUT = 13
    DELETED_TEXT = 14
    NEW_TEXT = 15
    NEW_RECIPIENT = 16
    SUB_RECIPIENT = 17
    NEW_MEMBERSHIP = 18
    NEW_USER_AREA = 19
    NEW_PRESENTATION = 20
    NEW_MOTD = 21
    TEXT_AUX_CHANGED = 22


#
# Classes for asynchronous messages from the server are all
# subclasses of AsyncMessage.
#

class AsyncMessage(object):
    MSG_NO = None

    def to_json(self):
        """Serializes the message to a dictionary that can be
        represented as json. The MSG_NO for the AsyncMessage should be
        available as the value for the key 'msg_no'.

        Subclasses should override this if they have more
        properties. The 'msg_no' key should still always be available.
        
        @returns A dictionary that can be json serialized.
        """
        return dict(msg_no=self.MSG_NO)

    @classmethod
    def parse(cls, conn):
        """Parses the message from the connection and returns a new
        instance of the parsed message."""
        raise NotImplementedError()

    @classmethod
    def from_json(cls, json_obj):
        """Parses the message from a dictionary that has be
        deserialized form json, and returns a new instance of the
        parsed message.
        
        @param json_obj A dictionary with the json representation.

        @returns A new instance of the message.
        """
        raise NotImplementedError()

# async-new-text-old [0] (1) Obsolete (10) <DEFAULT>
class AsyncNewTextOld(AsyncMessage):
    MSG_NO = AsyncMessages.NEW_TEXT_OLD

    def __init__(self, text_no, text_stat):
        self.text_no = text_no
        self.text_stat = text_stat

    def to_json(self):
        d = AsyncMessage.to_json(self)
        d['text_no'] = self.text_no
        # TODO: text_stat=self.text_stat.to_json()
        return d

    def __str__(self):
        return "<AsyncNewTextOld %d>" % (self.text_no,)

    @classmethod
    def parse(cls, conn):
        obj = cls(TextNo.parse(conn), TextStat.parse(conn, old_format=1))
        return obj

    @classmethod
    def from_json(cls, json_obj):
        obj = cls(json_obj['text_no'])
        #TODO: cls.text_stat = TextStat.from_json(json_obj['text_stat'])
        return obj

# async-i-am-off [1] (1) Obsolete
# async-i-am-on-onsolete [2] (1) Obsolete

# async-new-name [5] (1) Recommended <DEFAULT>
class AsyncNewName(AsyncMessage):
    MSG_NO = AsyncMessages.NEW_NAME
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.conf_no = ConfNo.parse(conn)
        obj.old_name = String.parse(conn)
        obj.new_name = String.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncNewName %d: %s => %s>" % (self.conf_no, self.old_name, self.new_name)

# async-i-am-on [6] Recommended
class AsyncIAmOn(AsyncMessage):
    MSG_NO = AsyncMessages.I_AM_ON
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.info = WhoInfo.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncIAmOn %s>" % (self.info, )

# async-sync-db [7] (1) Recommended <DEFAULT>
class AsyncSyncDB(AsyncMessage):
    MSG_NO = AsyncMessages.SYNC_DB
    @classmethod
    def parse(cls, conn):
        obj = cls()
        return obj

    def __str__(self):
        return "<AsyncSyncDB>"

# async-leave-conf [8] (1) Recommended <DEFAULT>
class AsyncLeaveConf(AsyncMessage):
    MSG_NO = AsyncMessages.LEAVE_CONF
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.conf_no = ConfNo.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncLeaveConf %d>" % (self.conf_no, )

# async-login [9] (1) Recommended <DEFAULT>
class AsyncLogin(AsyncMessage):
    MSG_NO = AsyncMessages.LOGIN

    def __init__(self, person_no, session_no):
        self.person_no = person_no
        self.session_no = session_no

    def to_json(self):
        d = AsyncMessage.to_json(self)
        d['person_no'] = self.person_no
        d['session_no'] = self.session_no
        return d

    def __str__(self):
        return "<AsyncLogin %d (session: %d)>" % (self.person_no, self.session_no)

    @classmethod
    def parse(cls, conn):
        obj = cls(PersNo.parse(conn), SessionNo.parse(conn))
        return obj

    @classmethod
    def from_json(cls, json_obj):
        obj = cls(json_obj['person_no'], json_obj['session_no'])
        return obj


# async-broadcast [10] Obsolete

# async-rejected-connection [11] (1) Recommended <DEFAULT>
class AsyncRejectedConnection(AsyncMessage):
    MSG_NO = AsyncMessages.REJECTED_CONNECTION
    @classmethod
    def parse(cls, conn):
        obj = cls()
        return obj

    def __str__(self):
        return "<AsyncRejectedConnection>"

# async-send-message [12] (1) Recommended <DEFAULT>
class AsyncSendMessage(AsyncMessage):
    MSG_NO = AsyncMessages.SEND_MESSAGE
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.recipient = ConfNo.parse(conn)
        obj.sender = PersNo.parse(conn)
        obj.message = String.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncSendMessage %d => %d: %s>" % (self.sender, self.recipient, self.message)

# async-logout [13] (1) Recommended <DEFAULT>
class AsyncLogout(AsyncMessage):
    MSG_NO = AsyncMessages.LOGOUT

    def __init__(self, person_no, session_no):
        self.person_no = person_no
        self.session_no = session_no

    def to_json(self):
        d = AsyncMessage.to_json(self)
        d['person_no'] = self.person_no
        d['session_no'] = self.session_no
        return d

    def __str__(self):
        return "<AsyncLogout %d (session: %s)>" % (self.person_no, self.session_no)

    @classmethod
    def parse(cls, conn):
        obj = cls(PersNo.parse(conn), SessionNo.parse(conn))
        return obj

    @classmethod
    def from_json(cls, json_obj):
        obj = cls(json_obj['person_no'], json_obj['session_no'])
        return obj

# async-deleted-text [14] (10) Recommended
class AsyncDeletedText(AsyncMessage):
    MSG_NO = AsyncMessages.DELETED_TEXT
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.text_no = TextNo.parse(conn)
        obj.text_stat = TextStat.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncDeletedText %d>" % (self.text_no,)

# async-new-text [15] (10) Recommended
class AsyncNewText(AsyncMessage):
    MSG_NO = AsyncMessages.NEW_TEXT

    def __init__(self, text_no, text_stat):
        self.text_no = text_no
        self.text_stat = text_stat

    def to_json(self):
        d = AsyncMessage.to_json(self)
        d['text_no'] = self.text_no
        # TODO: d['text_stat'] = self.text_stat.to_json()
        return d

    def __str__(self):
        return "<AsyncNewText %d>" % (self.text_no,)

    @classmethod
    def parse(cls, conn):
        obj = cls(TextNo.parse(conn), TextStat.parse(conn))
        return obj

    @classmethod
    def from_json(cls, json_obj):
        obj = cls(json_obj['text_no'], None)
        #TODO: cls.text_stat = TextStat.from_json(json_obj['text_stat'])
        return obj

# async-new-recipient [16] (10) Recommended
class AsyncNewRecipient(AsyncMessage):
    MSG_NO = AsyncMessages.NEW_RECIPIENT
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.text_no = TextNo.parse(conn)
        obj.conf_no = ConfNo.parse(conn)
        obj.type = Int32.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncNewRecipient %d: %d>" % (self.text_no, self.conf_no)

# async-sub-recipient [17] (10) Recommended
class AsyncSubRecipient(AsyncMessage):
    MSG_NO = AsyncMessages.SUB_RECIPIENT
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.text_no = TextNo.parse(conn)
        obj.conf_no = ConfNo.parse(conn)
        obj.type = Int32.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncSubRecipient %d: %d>" % (self.text_no, self.conf_no)

# async-new-membership [18] (10) Recommended
class AsyncNewMembership(AsyncMessage):
    MSG_NO = AsyncMessages.NEW_MEMBERSHIP
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.person_no = PersNo.parse(conn)
        obj.conf_no = ConfNo.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncNewMembership %d: %d>" % (self.person_no, self.conf_no)

# async-new-user-area [19] (11) Recommended
class AsyncNewUserArea(AsyncMessage):
    MSG_NO = AsyncMessages.NEW_USER_AREA
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.person_no = PersNo.parse(conn)
        obj.old_user_area = TextNo.parse(conn)
        obj.new_user_area = TextNo.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncNewUserArea %d: %d => %d>" % (
            self.person_no, self.old_user_area, self.new_user_area)

# async-new-presentation [20] (11) Recommended
class AsyncNewPresentation(AsyncMessage):
    MSG_NO = AsyncMessages.NEW_PRESENTATION
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.conf_no = ConfNo.parse(conn)
        obj.old_presentation = TextNo.parse(conn)
        obj.new_presentation = TextNo.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncNewPresentation %d: %d => %d>" % (
            self.conf_no, self.old_presentation, self.new_presentation)

# async-new-motd [21] (11) Recommended
class AsyncNewMotd(AsyncMessage):
    MSG_NO = AsyncMessages.NEW_MOTD
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.conf_no = ConfNo.parse(conn)
        obj.old_motd = TextNo.parse(conn)
        obj.new_motd = TextNo.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncNewMotd %d: %d => %d>" % (
            self.conf_no, self.old_motd, self.new_motd)

# async-text-aux-changed [22] (11) Recommended
class AsyncTextAuxChanged(AsyncMessage):
    MSG_NO = AsyncMessages.TEXT_AUX_CHANGED
    @classmethod
    def parse(cls, conn):
        obj = cls()
        obj.text_no = TextNo.parse(conn)
        obj.deleted = ArrayAuxItem.parse(conn)
        obj.added = ArrayAuxItem.parse(conn)
        return obj

    def __str__(self):
        return "<AsyncTextAuxChanged %d>" % (self.text_no,)


async_dict = {
    AsyncMessages.NEW_TEXT_OLD: AsyncNewTextOld,
    AsyncMessages.NEW_NAME: AsyncNewName,
    AsyncMessages.I_AM_ON: AsyncIAmOn,
    AsyncMessages.SYNC_DB: AsyncSyncDB,
    AsyncMessages.LEAVE_CONF: AsyncLeaveConf,
    AsyncMessages.LOGIN: AsyncLogin,
    AsyncMessages.REJECTED_CONNECTION: AsyncRejectedConnection,
    AsyncMessages.SEND_MESSAGE: AsyncSendMessage,
    AsyncMessages.LOGOUT: AsyncLogout,
    AsyncMessages.DELETED_TEXT: AsyncDeletedText,
    AsyncMessages.NEW_TEXT: AsyncNewText,
    AsyncMessages.NEW_RECIPIENT: AsyncNewRecipient,
    AsyncMessages.SUB_RECIPIENT: AsyncSubRecipient,
    AsyncMessages.NEW_MEMBERSHIP: AsyncNewMembership,
    AsyncMessages.NEW_USER_AREA: AsyncNewUserArea,
    AsyncMessages.NEW_PRESENTATION: AsyncNewPresentation,
    AsyncMessages.NEW_MOTD: AsyncNewMotd,
    AsyncMessages.TEXT_AUX_CHANGED: AsyncTextAuxChanged,
    }



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='pylyskom',
    version='0.2',
    description='Python LysKOM library',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Oskar Skoog',
    author_email='oskar@osd.se',
    url='https://github.com/osks/pylyskom',
    packages=['pylyskom'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: OS Independent',
    ],
    #python_requires='>=2.7, >=3.7',
    install_requires=[
        'six',
    ]
)



# -*- coding: utf-8 -*-
# LysKOM Protocol A version 10/11 client interface for Python
# (C) 1999-2002 Kent Engström. Released under GPL.
# (C) 2008 Henrik Rindlöw. Released under GPL.
# (C) 2012-2014 Oskar Skoog. Released under GPL.


# All errors belong to this class
class Error(Exception): pass

# All Protocol A errors are subclasses of ServerError
class ServerError(Error): pass
class NotImplemented(ServerError): pass # (2)
class ObsoleteCall(ServerError): pass # (3)
class InvalidPassword(ServerError): pass # (4)
class StringTooLong(ServerError): pass # (5)
class LoginFirst(ServerError): pass # (6)
class LoginDisallowed(ServerError): pass # (7)
class ConferenceZero(ServerError): pass # (8)
class UndefinedConference(ServerError): pass # (9)
class UndefinedPerson(ServerError): pass # (10)
class AccessDenied(ServerError): pass # (11)
class PermissionDenied(ServerError): pass # (12)
class NotMember(ServerError): pass # (13)
class NoSuchText(ServerError): pass # (14)
class TextZero(ServerError): pass # (15)
class NoSuchLocalText(ServerError): pass # (16)
class LocalTextZero(ServerError): pass # (17)
class BadName(ServerError): pass # (18)
class IndexOutOfRange(ServerError): pass # (19)
class ConferenceExists(ServerError): pass # (20)
class PersonExists(ServerError): pass # (21)
class SecretPublic(ServerError): pass # (22)
class Letterbox(ServerError): pass # (23)
class LdbError(ServerError): pass # (24)
class IllegalMisc(ServerError): pass # (25)
class IllegalInfoType(ServerError): pass # (26)
class AlreadyRecipient(ServerError): pass # (27)
class AlreadyComment(ServerError): pass # (28)
class AlreadyFootnote(ServerError): pass # (29)
class NotRecipient(ServerError): pass # (30)
class NotComment(ServerError): pass # (31)
class NotFootnote(ServerError): pass # (32)
class RecipientLimit(ServerError): pass # (33)
class CommentLimit(ServerError): pass # (34)
class FootnoteLimit(ServerError): pass # (35)
class MarkLimit(ServerError): pass # (36)
class NotAuthor(ServerError): pass # (37)
class NoConnect(ServerError): pass # (38)
class OutOfmemory(ServerError): pass # (39)
class ServerIsCrazy(ServerError): pass # (40)
class ClientIsCrazy(ServerError): pass # (41)
class UndefinedSession(ServerError): pass # (42)
class RegexpError(ServerError): pass # (43)
class NotMarked(ServerError): pass # (44)
class TemporaryFailure(ServerError): pass # (45)
class LongArray(ServerError): pass # (46)
class AnonymousRejected(ServerError): pass # (47)
class IllegalAuxItem(ServerError): pass # (48)
class AuxItemPermission(ServerError): pass # (49)
class UnknownAsync(ServerError): pass # (50)
class InternalError(ServerError): pass # (51)
class FeatureDisabled(ServerError): pass # (52)
class MessageNotSent(ServerError): pass # (53)
class InvalidMembershipType(ServerError): pass # (54)
class InvalidRange(ServerError): pass # (55)
class InvalidRangeList(ServerError): pass # (56)
class UndefinedMeasurement(ServerError): pass # (57)
class PriorityDenied(ServerError): pass # (58)
class WeightDenied(ServerError): pass # (59)
class WeightZero(ServerError): pass # (60)
class BadBool(ServerError): pass # (61)

# Mapping from Protocol A error_no to Python exception
error_dict = {
    2: NotImplemented,
    3: ObsoleteCall,
    4: InvalidPassword,
    5: StringTooLong,
    6: LoginFirst,
    7: LoginDisallowed,
    8: ConferenceZero,
    9: UndefinedConference,
    10: UndefinedPerson,
    11: AccessDenied,
    12: PermissionDenied,
    13: NotMember,
    14: NoSuchText,
    15: TextZero,
    16: NoSuchLocalText,
    17: LocalTextZero,
    18: BadName,
    19: IndexOutOfRange,
    20: ConferenceExists,
    21: PersonExists,
    22: SecretPublic,
    23: Letterbox,
    24: LdbError,
    25: IllegalMisc,
    26: IllegalInfoType,
    27: AlreadyRecipient,
    28: AlreadyComment,
    29: AlreadyFootnote,
    30: NotRecipient,
    31: NotComment,
    32: NotFootnote,
    33: RecipientLimit,
    34: CommentLimit,
    35: FootnoteLimit,
    36: MarkLimit,
    37: NotAuthor,
    38: NoConnect,
    39: OutOfmemory,
    40: ServerIsCrazy,
    41: ClientIsCrazy,
    42: UndefinedSession,
    43: RegexpError,
    44: NotMarked,
    45: TemporaryFailure,
    46: LongArray,
    47: AnonymousRejected,
    48: IllegalAuxItem,
    49: AuxItemPermission,
    50: UnknownAsync,
    51: InternalError,
    52: FeatureDisabled,
    53: MessageNotSent,
    54: InvalidMembershipType,
    55: InvalidRange,
    56: InvalidRangeList,
    57: UndefinedMeasurement,
    58: PriorityDenied,
    59: WeightDenied,
    60: WeightZero,
    61: BadBool,
    }

# All local errors are subclasses of LocalError
class LocalError(Error): pass
class BadInitialResponse(LocalError): pass # Not "LysKOM\n"
class BadRequestId(LocalError): pass  # Bad request id encountered
class ProtocolError(LocalError): pass # E.g. unexpected response
class UnimplementedAsync(LocalError): pass # Unknown asynchronous message
class ReceiveError(LocalError): pass # Error reading data from the server



import errno
import logging
import socket
import threading
import time


log = logging.getLogger('pylyskom.stats')


class Stats(object):
    def __init__(self, prefix=None):
        self._lock = threading.RLock()
        self._stats = dict()
        self._prefix = prefix

    def set(self, name, value, agg=None):
        assert agg is not None
        if self._prefix:
            name = self._prefix + name
        with self._lock:
            prev_value = self._stats.get(name)
            if prev_value:
                self._stats[name] = Stats._agg(agg, prev_value, value)
            else:
                self._stats[name] = value

    def dump(self):
        with self._lock:
            return self._stats.copy()

    def reset(self):
        with self._lock:
            old = self._stats
            self._stats = dict()
            return old

    @staticmethod
    def _agg(func, val1, val2):
        if func == 'last':
            return val2
        if func == 'sum':
            return val1 + val2
        elif func == 'min':
            return min(val1, val2)
        elif func == 'max':
            return max(val1, val2)
        elif func == 'avg':
            return (val1 + val2) / 2
        else:
            raise RuntimeError("Unknown aggregation function: {}".format(func))


stats = Stats(prefix='pylyskom.')


class GraphiteTcpConnection(object):
    def __init__(self, host, port=2003, timeout=None):
        self._host = host
        self._port = port
        self._timeout = timeout
        self._socket = None

    def connect(self):
        family, _, _, _, addr = socket.getaddrinfo(
            self._host, self._port, socket.AF_INET, socket.SOCK_STREAM)[0]
        self._socket = socket.socket(family, socket.SOCK_STREAM)
        self._socket.settimeout(self._timeout)
        self._socket.connect(addr)

    def close(self):
        if self._socket is None:
            return
        try:
            self._socket.close()
        except socket.error as e:
            if e.errno in (107, errno.ENOTCONN):
                # 107: Not connected anymore. Didn't find any errno
                # name, but the exception says "[Errno 107] Transport
                # endpoint is not connected".
                pass
            else:
                raise
        finally:
            self._socket = None

    def is_connected(self):
        if self._socket is None:
            return False
        return True

    def reconnect(self):
        self.close()
        self.connect()

    def send(self, metrics):
        if not self.is_connected():
            self.connect()
        data = serialize_graphite_plaintext(metrics)
        try:
            self._socket.sendall(data)
        except socket.error as e:
            # TODO: Attempt to reconnect and send again
            if e.errno in (107, errno.ENOTCONN):
                # 107: Not connected anymore. Didn't find any errno
                # name, but the exception says "[Errno 107] Transport
                # endpoint is not connected".
                self.close()
            elif e.errno in (32, errno.EPIPE): # Broken pipe
                self.close()
            else:
                raise


def serialize_graphite_plaintext(metrics):
    data = b''
    for m in metrics:
        # <metric path> <metric value> <metric timestamp>
        try:
            data += ("%s %s %d\n" % (m[0], m[1], m[2])).encode('ascii')
        except UnicodeDecodeError:
            log.exception("Failed to encode metrics as ascii: %s", m)
            # skip this metric
            continue
    return data


class StatsSender(threading.Thread):
    def __init__(self, statslist, conn, interval):
        threading.Thread.__init__(self)
        self._statslist = statslist
        self._conn = conn
        self._alive = threading.Event()
        self._alive.set()
        self._send_interval = interval

    def run(self):
        def step():
            self._step(time.time())
            return
        def should_continue():
            return self._alive.isSet()
        run_every(self._send_interval, step, should_continue)

    def join(self, timeout=None):
        self._alive.clear()
        threading.Thread.join(self, timeout)

    def _step(self, now):
        dump = merge_dicts([ s.dump() for s in self._statslist ])
        metrics = []
        for m in dump:
            metrics.append((m, dump[m], now))
        self._send(metrics)

    def _send(self, metrics):
        self._conn.send(metrics)


def run_every(interval, func, should_continue_func):
    sleep_max = 1.0
    init_t = time.time()
    init_iv = init_t // interval
    next_iv = init_iv + 1
    next_t = next_iv * interval
    while should_continue_func():
        now_t = time.time()
        now_iv = now_t // interval
        if now_t > next_t:
            func()
            after_t = time.time()
            after_iv = after_t // interval
            # skip interval if func() took more than interval time
            next_iv = max(now_iv, after_iv) + 1
            next_t = next_iv * interval
        after_t = time.time()
        sleep = min(sleep_max, max(0, next_t - after_t))
        time.sleep(sleep)


def merge_dicts(dicts):
    merged = dict()
    for d in dicts:
        merged.update(d)
    return merged
